#include "u.hpp"
#include "inter.hpp"
#include "ass.hpp"
#include "cp.hpp"
#include "known.hpp"

int verb=0;

void minmax(const cpl_t &cp,inter_t in[5])
{
  auto ex=cp.extr();
  int f=1;
  for(auto &it:ex) {
    for(int i=0;i<5;i++)
      if(f)
	in[i]=inter_t(it[i]);
      else
	in[i].add(it[i]);
    f=0;
  }
}

cpl_t basecpa() {  //returns \mathfrak{P}^\ge_\emptyset
  ass_t ca(5);
  vector<rat_t> v(5,1);
  ca.add(v,3);

  cpl_t cp(ca,1);

  for(int i=0;i<5;i++) { //each angle is in [0,1]
    vector<rat_t> v(5);
    v[i]=1;
    cp.cut(v,1);
    v[i]=-1;
    cp.cut(v,0);
  }

  for(int i=0;i<4;i++) { //angles are decreasing
    vector<rat_t> v(5);
    v[i]=-1;
    v[i+1]=1;
    cp.cut(v,0);
  }

  return cp;
}

cpl_t basecpu(){
  ass_t ca(5);
  rat_t v[5];
  for(int i=0;i<5;i++) v[i]=1;
  ca.add(v,0);

  return cpl_t(ca,1,1,1);
}

void compV(const vector<rat_t> &u,list<vi_t> &V,const vector<rat_t> &mina)
{
  for(int i=0;i<5;i++)
    if(mina[i]==0)
      assert(u[i]<0);

  vi_t v;
  for(v[0]=0;sumprod(v,mina,1)<=2;v[0]++)
    for(v[1]=0;sumprod(v,mina,2)<=2;v[1]++)
      for(v[2]=0;sumprod(v,mina,3)<=2;v[2]++)
	for(v[3]=0;sumprod(v,mina,4)<=2;v[3]++) {
	  if(u[3]<=0 && u[4]<=0 && sumprod(v,u,4)<0) break;
	  for(v[4]=0;sumprod(v,mina,5)<=2;v[4]++) {
	    if(u[4]<=0 && sumprod(v,u,5)<0) break;
	    if(sumprod(v,u)>=0)
	      V.push_back(v);
	  }
	}
}

vector<rat_t> findu0(const vvi_t &M,int k0)
{
  if(verb) printf("findu0 %d\n",k0);
  cpl_t cpu2=basecpu();

  for(auto &it : M) {
    vector<rat_t> vv(5);
    for(int i=0;i<5;i++)
      vv[i]=it[i];
    cpu2.proj(vv,0);
  }

  for(int i=0;i<5;i++)
    if(i>=5-k0) {
      vector<rat_t> vv(5);
      vv[i]=1;
      cpu2.cut(vv,0);
    }
  if(verb) cpu2.aff();
  auto s=cpu2.extr();

  int k=0;
  vector<rat_t> u(5);
  for(auto &it:s) {
    k++;
    for(int i=0;i<5;i++)
      u[i]+=it[i];
  }
  assert(k>0);
  for(int i=0;i<5;i++)
    u[i]/=k;

  return u;
}

vector<rat_t> findu0(const vector<rat_t> &mina,const vvi_t &M)
{
  auto u=findu0(M,2);
  if(mina[4]==0 && u[4]==0) u=findu0(M,1);
  return u;
}

bool isgood(const vvi_t &M,const cpl_t &cpu)
{
  auto s=cpu.extr();
  for(auto &it:s)
    for(auto &jt:M) {
      rat_t r=sumprod(jt,it);
      assert(r<=0);
      if(r<0) return 0;
    }
  return 1;
}

long long nbrec=0;
set<vvi_t> sols;

void recurse(const vvi_t &X,const cpl_t &cpa,const cpl_t &cpu,vvi_t notin)
{
  nbrec++;

  prank_t<rat_t> pp(6);
  pp.add(vector<int>(5,1),3);
  for(auto &it:X)
    pp.add(it,2);

  int rkB=pp.rank();

  if(verb) {
    printf("RECURSE cpa.d=%d cpu.d=%d rk=%d\n",cpa.dim(),cpu.dim(),rkB);
    for(int i=0;i<X.size();i++)      aff(X[i]);
  }

  if(cpa.empty()) {
    if(verb) printf("backtrack: P^\\ge_X empty\n");
    return;
  }

  inter_t in[5];
  vector<rat_t> mina(5);
  minmax(cpa,in);
  for(int i=0;i<5;i++) {
    if(in[i].a>=1 || in[i].b<=0) {
      if(verb) printf("backtrack: only degenerate solutions\n");
      return;
    }
    mina[i]=in[i].a;
    if(verb) cout<<"a_"<<i<<" >= "<<mina[i]<<endl;
  }

  vvi_t X3=compat(X);

  set<vi_t> sX;
  for(auto &it:X)
    sX.insert(it);

  auto cpu3=cpu;
  for(auto &it:X3)
    if(sX.find(it)==sX.end()) {
      vector<rat_t> vv(5);
      for(int i=0;i<5;i++)
	vv[i]=it[i];
      cpu3.cut(vv,0);
    }

  if(X3.size()>0 && isgood(X3,cpu3)) {
    sols.insert(sort(X3));
    printf("SOL ");
    for(auto &it:X3)      aff(it," ");
    printf("\n");
  }

  if(rkB==5) return;

  vector<rat_t> u=findu0(mina,X3);

  if(verb) aff(u,"\n","u=");

  list<vi_t> V;
  compV(u,V,mina);

  if(verb) {
    cout<<"V.size()="<<V.size()<<endl;
    for(auto &it:V) aff(it," ");
    printf("\n");
  }

  for(auto &v:V) {
    if(verb) aff(v,"...\n");

    bool ok=1;
    X3.push_back(v);
    if(!pp.add(v,2)) continue;

    for(auto &it:notin)
      if(pp.test(it,2)==0) {
	ok=0;
	break;
      }
    X3.pop_back();
    pp.del();
    if(!ok) continue;

    vector<rat_t> vv(5);
    for(int i=0;i<5;i++)
      vv[i]=v[i];

    auto cpa2=cpa;
    cpa2.proj(vv,2);
    if(cpa2.dim()<0)
      continue;

    if(!ok) continue;

    auto cpu2=cpu3;
    cpu2.cut(vv,0);

    X3.push_back(v);
    recurse(X3,cpa2,cpu2,notin);
    X3.pop_back();

    notin.push_back(v);
  }
}

int main(int ac, char **av)
{
  vector<rat_t> mina(5);
  inter_t in[5];
  minmax(basecpa(),in);
  for(int i=0;i<5;i++)
    mina[i]=in[i].a;

  verb=dbg_cp=0;
  while(ac>1 && av[1][0]=='-') {
    if(0==strcmp(av[1],"-v")) {
      verb++;
      dbg_cp++;
      ac--;av++;
      continue;
    }
    assert(0);
  }

  vvi_t X,notin;

  recurse(X,basecpa(),basecpu(),notin);
  cout<<"#rec : "<<nbrec<<endl;

  cout<<sols.size()<<" good sets when a_1>=...>=a_5\n";

  set<vvi_t> sols2,sols3;
  for(auto &vv:sols) {
    int p[5]={0,1,2,3,4};
    do {
      sols2.insert(sort(perm(vv,p)));
    } while(next_permutation(p,p+5));
  }
  cout<<sols2.size()<<" good sets\n";

  for(auto &vv:sols2)
    sols3.insert(minrep(vv));
  cout<<sols3.size()<<" good sets up to isomorphism\n";
  assert(sols3.size()==371);

  //order by dimension

  map<int,vector<vvi_t> > mo;
  for(auto &vv:sols3) {
    vector<vector<rat_t> > MM;
    vector<rat_t> v(5,1);
    ass_t ca(5);
    vector<int> v1(5,1);
    ca.add(v1,3);
    for(int i=0;i<vv.size();i++) {
      ca.add(vv[i],2);
    }
    int rk=5-ca.dim;
    mo[rk].push_back(vv);
  }

  FILE *out2=fopen("all.txt","w");

  int l=1;
  for(auto &it:mo) {
    for(auto &vv:it.second) {
      for(int z=0;z<vv.size();z++){
	for(int j=0;j<5;j++)
	  fprintf(out2,"%d ",vv[z][j]);
	fprintf(out2," ");
      }
      fprintf(out2,"\n");

      printf("    %3d: ",l);
      for(auto &it:vv) aff(it," ");
      printf("\n");

      auto c=sortsum(vv);
      auto r=rep(c);
      printf("rep %3d: ",l++);
      for(auto &it:r) aff(it," ");
      printf("\n");
    }
  }
  fclose(out2);

  return 0;
}
