#ifndef __CP_HPP__
#define __CP_HPP__

/* A class of a (bounded) convex polytope of R^5 */

#include "u.hpp"
#include "inter.hpp"
#include "ass.hpp"

int dbg_cp=-1;

template<class q_t,bool strict>
struct cp__t {
  struct hp_t {
    hp_t(){}
    hp_t(const vector<q_t> &v_,const q_t &q_):v(v_),q(q_) {}
    bool operator==(const hp_t &b) const { return v==b.v && q==b.q;}
    bool operator<(const hp_t &b) const { return make_pair(v,q)<make_pair(b.v,b.q);}
    vector<q_t> v;
    q_t q;
    void norm() {
      for(int i=0;i<v.size();i++) {
	if(v[i]!=q_t(0)) {
	  q_t a=q_t(1)/v[i];
	  if(a<0) a*=-1;
	  for(int i=0;i<v.size();i++)
	    v[i]*=a;
	  q*=a;
	  return;
	}
      }
      if(q>0) q=1;
      if(q<0) q=-1;
    }
    bool isnull() const {
      return ::isnull(v) && (q==0);
    }
  };
  struct cond_t {
    cond_t(){}
    cond_t(const hp_t &h5,const hp_t &hp):hp5(h5),hpp(hp){}
    hp_t hp5;
    hp_t hpp;
  };

  ass__t<q_t> c;
  set<vector<q_t> > s;
  vector<cond_t> conds;

  static q_t dot(const vector<q_t> &a, const vector<q_t> &b) {
    assert(a.size()==b.size());
    q_t r=0;
    for(int i=0;i<a.size();i++)
      r+=a[i]*b[i];
    return r;
  }

  static q_t dot(const q_t *a, const vector<q_t> &b) {
    q_t r=0;
    for(int i=0;i<b.size();i++)
      r+=a[i]*b[i];
    return r;
  }

  void kill() {
    c.dim=-1;
  }

  int dim() const {
    return c.dim;
  }

  bool empty() const {
    return c.dim<0;
  }

  cp__t():c(5) {
    vector<int> v1(5,1);
    c.add(v1,1);
    c.norm();

    for(int i=0;i<5;i++) {
      vector<q_t> v(5,0);
      v[i]=-1;
      if(!addcond(conds,hp_t(v,0))) kill();
    }

    recomp();
  }

  template<class q2_t>
  cp__t(const cp__t<q2_t,strict> &obj):c(obj.c) {
    for(auto &it:obj.s) {
      vector<q_t> p;
      p.resize(it.size());
      for(int i=0;i<p.size();i++)
	p[i]=q_t(it[i]);
      s.insert(p);
    }

    for(auto &it:obj.conds) {
      cond_t c;
      c.hp5.v.resize(it.hp5.v.size());
      for(int i=0;i<it.hp5.v.size();i++)
	c.hp5.v[i]=q_t(it.hp5.v[i]);
      c.hp5.q=q_t(it.hp5.q);
      c.hpp.v.resize(it.hpp.v.size());
      for(int i=0;i<it.hpp.v.size();i++)
	c.hpp.v[i]=q_t(it.hpp.v[i]);
      c.hpp.q=q_t(it.hpp.q);
      conds.push_back(c);
    }
  }

  template<class q2_t>
  cp__t(const ass__t<q2_t> &obj,int max1,const q_t &lim0=0,const q_t &lim1=1):c(obj) {
    for(int i=0;i<5;i++) {
      cond_t c;
      c.hp5.v.resize(5);
      c.hp5.v[i]=-1;
      c.hp5.q=lim0;
      proj(c.hp5,c.hpp);
      conds.push_back(c);
      if(max1) {
	c.hp5.v[i]=1;
	c.hp5.q=lim1;
	proj(c.hp5,c.hpp);
	conds.push_back(c);
      }
    }

    recomp();
  }

  bool operator<(const cp__t &b) const {
    if(c<b.c) return 1;
    if(b.c<c) return 0;
    return s<b.s;
  }

  set<vector<q_t> > extr() const {
    set<vector<q_t> > r;
    if(empty()) return r;
    for(auto &it:s)
      r.insert(c.proj(it));
    return r;
  }

  int test_(const vector<q_t> &v, const q_t &q) const {
    int x=0;
    for(auto &it:s) {
      auto r=dot(it,v);
      if(r>q) x|=1;
      else if(r<q) x|=4;
      else x|=2;
    }
    return x;
  }

  void proj(const hp_t &v5,hp_t &vp) const {
    if(dbg_cp>0) {cout<<"proj 5->deg from : ";::affd(v5.v," = ");cout<<getd(v5.q)<<endl;}
    if(empty()) return;
    vp.v.resize(c.dim);
    assert(v5.v.size()==5); //virer
    vp.q=v5.q-dot(c.o,v5.v);
    for(int i=0;i<c.dim;i++)
      vp.v[i]=dot(c.v[i],v5.v);
    if(dbg_cp>0) {cout<<"proj 5->deg to : ";::affd(vp.v," = ");cout<<getd(vp.q)<<endl;}
  }

  bool addcond(vector<cond_t> &conds,const hp_t &hp) {
    cond_t c;
    c.hp5=hp;
    proj(c.hp5,c.hpp);
    if(dbg_cp>1) {cout<<"add : ";::affd(c.hp5.v," ");cout<<getd(c.hp5.q)<<"  <-  ";::affd(c.hpp.v," ");cout<<getd(c.hpp.q)<<endl;}
    c.hpp.norm();
    if(::isnull(c.hpp.v)) {
      if(c.hpp.q>0) return 1;
      if(c.hpp.q<0) return 0;
      return !strict;
    }
    for(int i=0;i<conds.size();i++)
      if(conds[i].hpp==c.hpp) return 1;
    conds.push_back(c);
    return 1;
  }

  bool dimdim(hp_t &x) {
    for(auto &it:conds) {
      assert(!isnull(it.hpp.v));
      int r=test_(it.hpp.v,it.hpp.q);
      assert((r&1)==0);
      if((r&4)==0) {
	x=it.hp5;
	//printf("DIMDIM\n");
	return 1;
      }
    }
    return 0;
  }

  void recomp(const vector<cond_t> &conds,const ass__t<q_t> &c,int i,set<vector<q_t> > &ss,const set<int> &xx=set<int>()) {
    if(dbg_cp>1) {printf("cp:recomp i=%d deg=%d  ",i,c.dim);for(auto &it:xx) cout<<it<<".";cout<<endl;}
    if(empty()) return;
    if(c.dim==0) {
      vector<q_t> x(this->c.dim);
      for(int i=0;i<this->c.dim;i++)
	x[i]=c.o[i];
      if(dbg_cp>1) {cout<<"ep? ";::affd(x," .... ");}
      for(auto &it:conds) {
	auto r=dot(it.hpp.v,x);
	if(r>it.hpp.q) {
	  if(dbg_cp>1) {printf("out :");cout<<"r="<<r<<" > "<<it.hpp.q<<"   "; ::aff(it.hpp.v,"  ");	  cout<<it.hpp.q<<endl;}
	  return;
	}
      }
      if(dbg_cp>1) {printf("in\n");::aff(x);}
      ss.insert(x);
      return;
    }
    for(int u=i;u<conds.size();u++) {
      auto c2=c;
      c2.add(conds[u].hpp.v,conds[u].hpp.q);
      auto xx2=xx;
      xx2.insert(u);
      recomp(conds,c2,u+1,ss,xx2);
    }
  }

  void recomp_d() {
    if(empty()) return;
    if(dbg_cp>0) cout<<"recomp_d...\n";
    int d=c.dim;
    if(dbg_cp>0) {printf(" RN (d=%d)",c.dim);    affd();}
    vector<cond_t> conds2;
    for(int i=0;i<conds.size();i++) {
      if(dbg_cp>0) cout<<" || ";
      bool f=1;
      vector<vector<q_t> > M;
      vector<q_t> o;
      int p=0;
      int pp=0;
      for(auto &it:s) {
	auto r=dot(it,conds[i].hpp.v);
	auto q=conds[i].hpp.q;
	assert(r<=q);

	if(r==q) {
	  if(dbg_cp>0) cout<<pp<<" ";
	  p++;
	  //on a un point sur la face;
	  if(f) {
	    o=it;
	    f=0;
	  } else {
	    vector<q_t> v=it;
	    for(int i=0;i<v.size();i++)
	      v[i]-=o[i];
	    M.push_back(v);
	  }
	}
	pp++;
      }
      if(dbg_cp>0) cout<<" M.size="<<M.size();
      int rk=0;
      if(M.size()>=d-1) {
	rk=mrank(M,c.dim);
	if(dbg_cp>0) cout<<"("<<M.size()<<":"<<rk<<" d="<<d<<")";
	assert(rk<=d-1); // sinon, on serait pas sur un hyperplan...
      }
      if(p && rk==d-1) {
	conds2.push_back(conds[i]);
	if(dbg_cp>0) cout<<(" in ");
      } else {
	if(dbg_cp>0) cout<<(" out ");
      }
    }
    if(dbg_cp>0) printf("\n");
    conds.swap(conds2);
  }

  void recomp_s() {
    if(empty()) return;
    if(dbg_cp>0) cout<<"recomp_s...\n";
    s.clear();
    ass__t<q_t> c(this->c.dim);
    if(dbg_cp>1) c.aff();
    assert(conds.size()<30);
    recomp(conds,c,0,s);
  }

  void recomp() {
    recomp_s();
    recomp_d();
  }

  void cut_(const cond_t &co) {
    int r=test_(co.hpp.v,co.hpp.q);
    if(dbg_cp>0) printf("cut_ r=%d\n",r);
    if((r&6)==0) {
      kill();
      return;
    }
    if((r&4)==0 && strict) {
      kill(); //aucun point de l'autre coté, on devient dégénéré...
      return;
    }
    if((r&1)==0) //rien à faire
      return;

    int r2=addcond(conds,co.hp5);
    assert(r2);
    recomp_s();
    hp_t h;
    while(dimdim(h)) {
      if(dbg_cp>0) {printf("dimdim(cut) \n"); ::aff(h.v,"   ");cout<<h.q<<endl;}
      if(strict) {
	kill();
	return;
      }
      c.add(h.v,h.q);
      c.norm();
      assert(!empty()); // théoriquement, le polytope ne peut pas être vide
      recomp_hp();
      recomp_s();
    }
    recomp_d();
  }

public:

  void aff(ostream &cout=std::cout) const {
    cout<<"deg="<<c.dim<<endl;
    if(empty()) return;
    c.aff(cout);
    for(auto &it:s) {
      cout<<"EP ";
      ::aff(it," ->  ","",cout);
      vector<q_t> v=c.proj(it);
      ::aff(v,"\n","",cout);
    }
    for(auto &it:conds) {
      ::aff(it.hpp.v,"","",cout);
      cout<<" < "<<it.hpp.q<<"  <-  ";
      ::aff(it.hp5.v,"","",cout);
      cout<<" < "<<it.hp5.q<<endl;
    }
  }

  void affd(ostream &cout=std::cout) const {
    printf("deg=%d\n",c.dim);
    if(empty()) return;
    c.affd(cout);
    for(auto &it:s) {
      cout<<"EPd ";
      ::affd(it," ->  ","",cout);
      vector<q_t> v=c.proj(it);
      ::affd(v,"\n","",cout);
    }
    for(auto &it:conds) {
      ::affd(it.hpp.v,"","",cout);
      cout<<" < "<<getd(it.hpp.q)<<"  <-  ";
      ::affd(it.hp5.v,"","",cout);
      cout<<" < "<<getd(it.hp5.q)<<endl;
    }
  }

  int cut(const vector<q_t> &v, const q_t &q) {
    if(empty()) return 0;
    cond_t c;
    c.hp5.v=v;c.hp5.q=q;
    proj(c.hp5,c.hpp);
    if(dbg_cp>0) {::aff(c.hpp.v,"  ");cout<<c.hpp.q<<endl;}
    cut_(c);
    return !empty();
  }

  void recomp_hp() {
    if(dbg_cp>0) cout<<"recomp_hp...\n";
    if(empty()) return;
    map<hp_t,cond_t> bis;
    for(auto &it:conds) {
      hp_t b;
      proj(it.hp5,b);
      b.norm();
      if(::isnull(b.v)) {
	if(b.q<q_t(0) || (b.q==q_t(0) && strict)) {
	  kill();
	  return;
	}
	continue;
      }
      bis[b]=cond_t(it.hp5,b);
    }
    conds.clear();
    for(auto &it:bis)
      conds.push_back(it.second);
  }

  void proj(const vector<q_t> &v, const q_t &q) {
    if(dbg_cp>0) {printf("proj : ");::affd(v,"    =   ");cout<<getd(q)<<endl;}
    if(empty()) return;

    int d=c.dim;
    c.add(v,q);
    c.norm();
    if(dbg_cp) {printf("ass apres proj:\n");c.aff();}
    if(c.dim==d) {
      return;
    }
    if(empty()) return;
    recomp_hp();
    if(empty()) return;
    if(dbg_cp) {printf("apres recalcul hp:\n");aff();}
    recomp_s();
    if(s.size()==0) {
      kill();
      return;
    }
    assert(!empty()); //virer
    hp_t h;
    while(dimdim(h)) {
      if(strict) {
	kill();
	return;
      }
      c.add(h.v,h.q);
      c.norm();
      assert(c.dim>=0); //virer
      recomp_hp();
      assert(!empty()); //virer
      recomp_s();
      assert(!empty()); //virer
    }
    recomp_d();
  }

};

template<class q_t,bool str>
bool desc_(cp__t<q_t,str> &cl2,const q_t &mm) {
  if(cl2.empty()) return 0;

  auto ex=cl2.extr();
  int f=1;
  inter__t<q_t> in[5];
  for(auto &it:ex) {
    for(int i=0;i<5;i++) {
      if(f) {
	in[i]=inter__t<q_t>(it[i]);
      } else {
	in[i].add(it[i]);
      }
    }
    f=0;
  }
  if(f) return 0;

  q_t m=0;
  int j=-1;
  for(int i=0;i<5;i++)
    if(m<in[i].len()) m=in[j=i].len();
  if(j==-1 || m<mm) return 0;

  int d=cl2.c.dim;
  vector<q_t> v(5);
  if(dbg_cp) {cl2.affd();cout<<"desc " <<j<<" "<< getd(in[j].a) << " " <<getd(in[j].b)<<endl;}
  v[j]=1;
  cl2.proj(v,in[j].mid());
  if(dbg_cp) {printf("after proj:\n");cl2.affd();}
  assert(cl2.c.dim==d-1);
  return 1;
}

using cp_t = cp__t<rat_t,1> ;
using cpl_t = cp__t<rat_t,0> ;

bool desc(cp_t &cl2,const rat_t &mm) {
  return desc_<rat_t>(cl2,mm);
}

#endif
