#ifndef _EXACTDEGEN_HPP_
#define _EXACTDEGEN_HPP_

#include "cp.hpp"
#include "an.hpp"

map<int,poly_t> pT,pU;

//Chebyshev polynomials
poly_t T(int i) {
  assert(i>=0);
  if(i==0) return poly_t(1);
  if(i==1) return poly_t(1,1);
  auto it=pT.find(i);
  if(it!=pT.end()) return it->second;
  return pT[i]=poly_t(2,1)*T(i-1)-T(i-2);
}

poly_t U(int i) {
  assert(i>=-1);
  if(i==-1) return poly_t(0);
  if(i==0) return poly_t(1);
  if(i==1) return poly_t(2,1);
  auto it=pU.find(i);
  if(it!=pU.end()) return it->second;
  return pU[i]=poly_t(2,1)*U(i-1)-U(i-2);
}

poly_t Pc(int i) {
  if(i>=0) return T(i);
  return T(-i);
}

poly_t Ps(int i) {
  if(i>=0) return U(i-1);
  return -U(-i-1);
}

//minimal polynomials of cos(pi/q), with a bound of the root cos(pi/q) between b/100 and 1
int tminpol[]={
  // q, b, poly
  3, 10,  0,-1,  1,2,  -1,
  4, 50,  0,-1,  2,2,  -1,
  5, 60,  0,-1,  1,-2,  2,4,  -1,
  6, 80,  0,-3,  2,4,  -1,
  7, 80,  0,1,  1,-4,  2,-4,  3,8,  -1,
  8, 90,  0,1,  2,-8,  4,8,  -1,
  9, 93,  0,-1,  1,-6,  3,8,  -1,
  10, 94,  0,5,  2,-20,  4,16,  -1,
  11, 94,  0,-1,  1,6,  2,12,  3,-32,  4,-16,  5,32,  -1,
  12, 95,  0,1,  2,-16,  4,16,  -1,
  13, 96,  0,-1,  1,-6,  2,24,  3,32,  4,-80,  5,-32,  6,64,  -1,
  15, 96,  0,1,  1,-8,  2,-16,  3,8,  4,16,  -1,
  16, 97,  0,1,  2,-32,  4,160,  6,-256,  8,128,  -1,
  17, 97,  0,1,  1,8,  2,-40,  3,-80,  4,240,  5,192,  6,-448,  7,-128,  8,256,  -1,
  18, 97,  0,-3,  2,36,  4,-96,  6,64,  -1,
  19, 97,  0,-1,  1,10,  2,40,  3,-160,  4,-240,  5,672,  6,448,  7,-1024,  8,-256,  9,512,  -1,
  21, 97,  0,1,  1,16,  2,32,  3,-48,  4,-96,  5,32,  6,64,  -1,
  24, 98,  0,1,  2,-64,  4,320,  6,-512,  8,256,  -1,
  27, 98,  0,-1,  1,18,  3,-240,  5,864,  7,-1152,  9,512,  -1,
  -1
};

map<int,root_t> minpol;

void initminpol()
{
  int i=0;
  while(tminpol[i]>=0) {
    poly_t t;
    int n=tminpol[i++];
    int br=tminpol[i++];
    while(tminpol[i]>=0) {
      t+=poly_t(tminpol[i+1],tminpol[i]);
      i+=2;
    }
    i++;
    minpol[n]=root_t(t,rat_t(br)/100,1);
  }
}

int initminpol_=(initminpol(),0);

int dbg_cpan=-1;

void makean(const ass_t &ca,vector<an_t> &psv,vector<an_t> &pcv)
{
  auto sa=ca;
  sa.sum();

  mpz_class pp_=1;
  vector<int> a(5);
  for(int i=0;i<5;i++)
    pp_=ppcm(pp_,mpz_class(sa.o[i].get_den()));

  for(int i=0;i<5;i++) {
    rat_t t=sa.o[i]*pp_;
    assert(t.get_den()==1);
    a[i]=t.get_num().get_si();
  }
  int pp=pp_.get_si();
  long double aa=cosl(M_PI/pp);

  if(dbg_cpan) {printf("ca:\n");ca.aff();printf("sa:\n");sa.aff();printf("angle rat: ");::aff(a,"");printf(" / %d\n",pp);}

  root_t root(minpol.at(pp));
  assert(fabs(root.p.eval(aa)<0.0001));
  if(dbg_cpan>1) cout<<"poly="<<root.p<<endl;

  psv=pcv=vector<an_t>(5,an_t(poly_t(0),root));

  for(int i=0;i<5;i++) {
    auto c=Pc(a[i]);
    if(a[i]<20) //overflows
      assert(fabs(c.eval(aa)-cosl(a[i]*M_PI/pp))<0.0001);
    pcv[i]+=c;
  }

  for(int i=0;i<5;i++) {
    auto c=Ps(a[i]);
    if(a[i]<20)
      assert(fabs(c.eval(aa)*sinl(M_PI/pp)-sinl(a[i]*M_PI/pp))<0.0001);
    psv[i]+=c;
  }
}

bool cpan(const ass_t &ca,const cp_t &cp,vector<double> &vs, vector<double> &vs2) // -1: sait pas  0: degenere 1: pas degenere
{
  assert(ca.dim==0);
  if(dbg_cpan) printf("exact computation on AN\n");

  cp__t<an_t,1> cpb(cp);
  if(dbg_cpan) {printf("cpb:\n");cpb.affd();}

  auto sa=ca;
  sa.sum();

  vector<an_t> psv,pcv;
  makean(ca,psv,pcv);
  cpb.proj(pcv,0);
  cpb.proj(psv,0);

  if(dbg_cpan) {printf("cpb : \n");cpb.aff();}

  if(cpb.empty()) return 0;
  //if(cpb.faced) return 0;

  while(desc_(cpb,an_t(0)));
  assert(cpb.dim()==0); // && cpb.faced==0

  auto s=cpb.extr();
  assert(!s.empty());
  auto p=*(s.begin());
  vs.resize(5);
  for(int i=0;i<5;i++)
    vs[i]=getd(ca.o[i]);
  vs2.resize(5);
  for(int i=0;i<5;i++)
    vs2[i]=getd(p[i]);

  return 1;
}

vi_t findvect(const ass_t &ca,const ass_t &cb)
{
  if(ca.dim!=0) return v0;
  if(dbg_cpan) printf("findvect\n");

  ass__t<an_t> cb2(cb);

  vector<an_t> psv,pcv;
  makean(ca,psv,pcv);

  cb2.add(pcv,0);
  cb2.add(psv,0);

  if(cb2.dim<0) {
    return v0;
  }

  for(int p=1;p<=2;p++) {
    for(int i=0;;i++) {
      vi_t v(v0);
      int j=i;
      for(int k=0;k<5;k++) {
	v[k]=(j%(2*p+1))-p;
	j/=(2*p+1);
      }
      if(j) break;
      if(!cb.comp(v,0) && cb2.comp(v,an_t(0))) {
	if(dbg_cpan) ::aff(v,"\n","-> ");
	return v;
      }
    }
  }

  if(dbg_cpan) cout<<"nothing!\n";

  return v0;
}

#endif
