#ifndef _INTER_HPP_
#define _INTER_HPP_

#include <iostream>
#include <assert.h>

using namespace std;

template<class q_t>
struct inter__t {
  q_t a,b; //we have always a<=b

  inter__t(const q_t &x=0):a(x),b(x) {}

  inter__t(const q_t &x,const q_t &y):a(x),b(y) {}

  static q_t min(const q_t &x,const q_t &y) {if(x<y) return x; return y;}
  static q_t max(const q_t &x,const q_t &y) {if(x>y) return x; return y;}

  inter__t(const inter__t &x,const inter__t &y):a(min(x.a,y.a)),b(max(x.b,y.b)) {}

  bool cont0() const {
    return a<=0 && b>=0;
  }

  void add(const q_t &x) {
    if(a>x) a=x;
    if(b<x) b=x;
  }

  q_t len() const {
    return b-a;
  }

  q_t mid() const {
    return (a+b)/2;
  }

  inter__t operator+(const inter__t &i) const {
    return inter__t(a+i.a,b+i.b);
  }

  inter__t operator-() const {
    return inter__t(-b,-a);
  }

  inter__t operator-(const inter__t &i) const {
    return inter__t(a-i.b,b-i.a);
  }

  inter__t operator*(const q_t &m) const {
    if(m>=0)
      return inter__t(a*m,b*m);
    return inter__t(b*m,a*m);
  }

  bool operator==(const inter__t &o) const { return a==o.a && b==o.b;}

};

using inter_t=inter__t<rat_t>;

inter_t iatan(const rat_t &x,const rat_t &pr) // compute an interval for atan(x) with precision pr (only used to compute an interval for pi)
{
  rat_t r=0;
  rat_t xp=1;
  assert(x>0 && x<1);
  for(int i=0;;i++) {
    xp*=x;
    r+=xp/(4*i+1);
    xp*=x*x;
    r-=xp/(4*i+3);
    xp*=x;
    if(xp/(1-x)<pr)
      break;
  }
  return inter_t(r,r+pr);
}

inter_t ipi(const rat_t &pr) // compute an interval for pi precision pr, with the formula pi/4=4*atan(1/5)-\arctan(1/239)
{
  return (iatan(rat_t(1)/5,pr/32)*4-iatan(rat_t(1)/239,pr/8))*4;
}

pair<inter_t,inter_t> isincos_(const rat_t &x,const rat_t &pr) // compute an interval for both sin(x) and cos(x) (x must be in ]-1..1[)
{
  rat_t c=0,s=0;

  rat_t f=1;
  rat_t xp=1;
  assert(x>-1 && x<1);
  for(int i=0;;i+=4) {
    c+=xp/f;
    xp*=x;f*=i+1;
    s+=xp/f;
    xp*=x;f*=i+2;
    c-=xp/f;
    xp*=x;f*=i+3;
    s-=xp/f;
    xp*=x;f*=i+4;
    if(abs(xp)/(1-abs(x))<f*pr/2)
      break;
  }
  return make_pair(inter_t(s-pr/2,s+pr/2),inter_t(c-pr/2,c+pr/2));
}

pair<inter_t,inter_t> isincospi_(const rat_t &x,const rat_t &pr) // compute an interval for both sin(x*pi) and cos(x*pi)
{
  rat_t bb=pr;
  while(1) {
    bb/=2;
    inter_t aa=ipi(bb)*x;
    auto ra=isincos_(aa.a,bb);
    auto rb=isincos_(aa.b,bb);
    inter_t rs(ra.first,rb.first);
    inter_t rc(ra.second,rb.second);
    if(aa.cont0()) {
      rc.add(1);
      rs.add(0);
    }
    if(rs.len()>pr) continue;
    if(rc.len()>pr) continue;
    return make_pair(rs,rc);
  }
}

pair<inter_t,inter_t> isincospi(rat_t x,const rat_t &pr) // compute an interval for both sin(x*pi) and cos(x*pi) (no condition for x)
{
  rat_t h(1);h/=2;
  rat_t q=h/2;
  while(x>2-q) x-=2;
  while(x<-q) x+=2;
  int d=0;
  while(x>q) {
    x-=h;
    d++;
  }
  auto r=isincospi_(x,pr);
  if(d==0) return r;
  if(d==1) return make_pair(r.second,-r.first);
  if(d==2) return make_pair(-r.first,-r.second);
  assert(d==3);
  return make_pair(-r.second,r.first);
}

pair<inter_t,inter_t> isincospi(inter_t i,const rat_t &pr) // compute an interval for both sin(x*pi) and cos(x*pi), for all x in i
{
  while(i.a<0) {    i.a+=2;    i.b+=2;  }
  while(i.a>2) {    i.a-=2;    i.b-=2;  }
  auto ra=isincospi(i.a,pr);
  auto rb=isincospi(i.b,pr);
  inter_t rs(ra.first,rb.first);
  inter_t rc(ra.second,rb.second);
  for(int z=0;z<2*i.b;z++) {
    if(2*i.a<z) {
      if(z%4==0) { rc.add(1);rs.add(0);}
      if(z%4==1) { rc.add(0);rs.add(1);}
      if(z%4==2) { rc.add(-1);rs.add(0);}
      if(z%4==3) { rc.add(0);rs.add(-1);}
    }
  }
  return make_pair(rs,rc);
}

#endif
