#ifndef _KNOWN_HPP_
#define _KNOWN_HPP_

#include "ass.hpp"
#include "cp.hpp"

vector<ass_t> knowna,knownl;
vector<vvi_t> cknowna,cknownl;

bool isiso(const ass_t &a_,const ass_t &b_,ass_t &aa,ass_t &bb,bool skipb=0)
{
  auto a=a_;
  auto b=b_;
  a.norm();
  b.norm();
  aa.norm();
  bb.norm();
  for(int i=0;i<5;i++) {
    if(a==aa && (skipb || b==bb)) return 1;
    aa.perm(cperm);
    bb.perm(cperm);
    aa.norm();
    bb.norm();
  }

  aa.perm(mperm);
  bb.perm(mperm);
  bb.perm(cperm);
  aa.norm();
  bb.norm();

  for(int i=0;i<5;i++) {
    if(a==aa && (skipb || b==bb)) return 1;
    aa.perm(cperm);
    bb.perm(cperm);
    aa.norm();
    bb.norm();
  }

  return 0;
}

set<int> isexactknown(const ass_t &a,const ass_t &b)
{
  set<int> r;
  for(int j=0;j<knowna.size();j++) {
    ass_t aa=knowna[j];
    ass_t bb=knownl[j];
    if(isiso(a,b,aa,bb))
      r.insert(j);
  }
  return r;
}

set<int> isknown(const ass_t &a,const ass_t &b)
{
  set<int> r;
  for(int j=0;j<knowna.size();j++) {
    ass_t aa=a;
    ass_t bb=b;
    aa.norm();
    bb.norm();
    for(int i=0;i<5;i++) {
      if(aa.vcomp(cknowna[j],2) && bb.vcomp(cknownl[j],0)) {r.insert(j);}
      aa.perm(cperm);
      bb.perm(cperm);
      aa.norm();
      bb.norm();
    }

    aa.perm(mperm);
    bb.perm(mperm);
    bb.perm(cperm); // 1 ou 4 ? (1 marche, pas 4...)
    aa.norm();
    bb.norm();

    for(int i=0;i<5;i++) {
      if(aa.vcomp(cknowna[j],2) && bb.vcomp(cknownl[j],0)) {r.insert(j);}
      aa.perm(cperm);
      bb.perm(cperm);
      aa.norm();
      bb.norm();
    }
  }
  return r;
}

set<int> isexactknown(const ass_t &a)
{
  set<int> r;
  for(int j=0;j<knowna.size();j++) {
    ass_t aa=a;
    aa.norm();
    for(int i=0;i<5;i++) {
      if(knowna[j]==aa){r.insert(j);}
      aa.perm(cperm);
      aa.norm();
    }

    aa.perm(mperm);
    aa.norm();

    for(int i=0;i<5;i++) {
      if(knowna[j]==aa){r.insert(j);}
      aa.perm(cperm);
      aa.norm();
    }
  }
  return r;
}

set<int> isknown(const ass_t &a)
{
  set<int> r;
  for(int j=0;j<knowna.size();j++) {
    ass_t aa=a;
    aa.norm();
    for(int i=0;i<5;i++) {
      if(aa.vcomp(cknowna[j],2)) {r.insert(j);}
      aa.perm(cperm);
      aa.norm();
    }

    aa.perm(mperm);
    aa.norm();

    for(int i=0;i<5;i++) {
      if(aa.vcomp(cknowna[j],2)) {r.insert(j);}
      aa.perm(cperm);
      aa.norm();
    }
  }
  return r;
}

void allcomp(const vector<rat_t> &vs,vvi_t &vc,vi_t &u,int i)
{
  rat_t tt=0;
  for(int j=0;j<i;j++)
    tt+=u[j]*vs[j];
  if(i==5) {
    if(tt==2)
      vc.push_back(u);
    return;
  }
  for(int z=0;;z++) {
    if(tt+z*vs[i]>2) return;
    u[i]=z;
    allcomp(vs,vc,u,i+1);
  }
}

void allcomp(const vector<rat_t> &vs,vvi_t &vc)
{
  vi_t u(v0);
  allcomp(vs,vc,u,0);
}

vvi_t compat(const ass_t &ca)
{
  cp_t cl(ca,1);
  for(int i=0;i<5;i++) {
    vector<rat_t> v(5);
    v[i]=-1;
    cl.cut(v,0);
  }

  while(desc(cl,0));
  assert(cl.dim()==0);

  vector<rat_t> vs;
  vs.resize(5);
  for(int i=0;i<5;i++)
    vs[i]=cl.c.o[i];

  vvi_t T;
  allcomp(vs,T);

  vvi_t X;
  for(auto &it:T) {
    if(ca.comp(it,2)) {
      X.push_back(it);
    }
  }

  return X;
}

vvi_t compat(const vvi_t &X)
{
  ass_t ca(5);
  vector<rat_t> v(5,1);
  ca.add(v,3);

  for(int j=0;j<X.size();j++) {
    for(int i=0;i<5;i++)
      v[i]=X[j][i];
    ca.add(v,2);
  }

  return compat(ca);
}

vvi_t rep(const vvi_t &M)
{
  vvi_t r;
  ass_t ca(5);
  vector<int> v1(5,1);
  ca.add(v1,3);
  for(int i=0;i<M.size();i++) {
    int d=ca.dim;
    ca.add(M[i],2);
    if(ca.dim<d)
      r.push_back(M[i]);
  }
  return r;
}

vi_t toint(const vector<rat_t> &v)
{
  mpz_class pp=1;
  for(int i=0;i<5;i++)
    pp=ppcm(pp,v[i].get_den());
  vi_t r;
  for(int i=0;i<5;i++) {
    rat_t x=v[i]*pp;
    assert(x.get_den()==1);
    r[i]=x.get_num().get_si();
  }
  return r;
}

void readknowntypes()
{
  FILE *in=fopen("types.txt","r");
  assert(in);
  if(in==NULL) return;
  char bf[1000];
  while(fgets(bf,999,in)) {
    printf("known2.txt: %s",bf);
    if(bf[0]!='A') continue;
    ass_t a(5);
    ass_t b(5);
    a.read(bf+2);
    a.norm();

    assert(a.comp(vector<int>(5,1),3));

    knowna.push_back(a);

    auto vvia=compat(a);
    cknowna.push_back(rep(vvia));

    auto r=fgets(bf,999,in);
    assert(r);
    printf("k: %s",bf);
    printf("(number %d)\n",int(knowna.size())-1);
    assert(bf[0]=='B');
    b.read(bf+2);
    b.norm();

    knownl.push_back(b);

    ass_t db(5),rb(5);
    db.add(b.o,0);
    for(int i=0;i<b.dim;i++)
      db.add(b.v[i],0);

    vvi_t vvib;

    vector<rat_t> v1(5,1);
    assert(isnull(rb.o));
    rb.add(v1,1);
    for(int i=0;i<db.dim;i++) {
      vi_t r=toint(db.v[i]);
      //::aff(r,"\n","B ");
      rb.add(r,0);
      vvib.push_back(r);
    }
    rb.norm();
    assert(b==rb);

    cknownl.push_back(vvib);

  }
  fclose(in);
}

unsigned long long isknownB(const ass_t &mainca,const ass_t &cl) {
  static map<ass_t,unsigned long long> mk;
  auto it=mk.find(cl);
  if(it!=mk.end()) return it->second;

  unsigned long long r=0;
  auto ek=isknown(mainca,cl);
  auto eek=isexactknown(mainca,cl);

  for(auto &it:ek) {
    r|=1LL<<it;
  }
  for(auto &it:eek) {
    r|=1LL<<(it+32);
    assert(ek.find(it)!=ek.end());
  }
  mk[cl]=r;
  return r;
}

vvi_t sortsum(const vvi_t &M,bool opt=0)
{
  map<int,set<vi_t> > s;
  for(auto &it:M) {
    auto v=it;
    if(opt && iseven(v)) v=div2(v);
    s[sum(v)].insert(it);
  }
  vvi_t r;
  for(auto &it:s)
    for(auto &jt:it.second)
      r.push_back(jt);
  return r;
}

#endif
