#include "u.hpp"
#include "cp.hpp"
#include "known.hpp"
#include "cpan.hpp"
#include "hassol.hpp"

#ifdef CAIRO
#include <cairo.h>
#endif

#define ANG_0 '0'
#define ANG_180 '_'
#define ANG_UNK '?'

int skipknown=0;
set<int> knowntoskip;

int algobnd=2;
int nocut=0;

int pause(const string &info);
struct t_t;
void settaff(const t_t &t);

ass_t mainca(5); //angle conditions (=\mathfrak{P}_\mathcal{X}) for all the process

vector<vi_t> vA2,vA1; // vA2=\mathcal{W}^c   vA1 is the set of possible vector types for half vertices

void setmainca(const ass_t &ca) {
  mainca=ca;
  mainca.norm();

  vA2=compat(ca);

  vA1.clear();
  for(auto &it:vA2)
    if(iseven(it)) {
      vi_t t=div2(it);
      vA1.push_back(t);
    }

  if(vA1.empty())
    printf("edge-to-edge\n");
  for(auto &it:vA1)
    for(int i=0;i<5;i++)
      if(it[i]==1 && it[(i+1)%5]==1 && sum(it)==2) printf("type1\n");
}

int dbg_nt=-1;

struct t_t {

  cp_t cl3;

  struct he_t {
    int point;
    int opp; //he opp
    int next; //he suivant dans la face;
    int prev;
    int face;
    int fsens; //1,-1, ou 0 si spécial //sert seulement pour calcul position

    char angle; //"a-eA-E" ou ANG_*
    double posa;

    he_t(){
      fsens=0;
      opp=point=next=face=-1;
      angle=ANG_UNK;
    }
  };

  list<int> freepoints;
  bool isfreepoint(int u) const {
    return vpoint[u].he<0;
  }
  struct point_t {
    int he;
    double posx,posy;
  };

  vector<he_t> vhe;
  vector<point_t> vpoint;
  int nbfaces;

  t_t() {
    nbfaces=0;
    solved=0;
    vs_.resize(5);
    vs2_.resize(5);
  }

  void makepenta(int sens=1,int *th=NULL)
  {
    int p[5],he[5],heo[5];
    int f=nbfaces;
    nbfaces++;
    for(int i=0;i<5;i++) {
      if(freepoints.empty()) {
	p[i]=vpoint.size();
	vpoint.push_back(point_t());
      } else {
	p[i]=freepoints.front();
	freepoints.pop_front();
      }
      he[i]=vhe.size();vhe.push_back(he_t());
      heo[i]=vhe.size();vhe.push_back(he_t());
      vhe[he[i]].point=p[i];
      vhe[he[i]].face=f;
      vhe[he[i]].fsens=sens;
      vhe[he[i]].angle=((sens*i+10)%5)+(sens==1?'a':'A');
      vpoint[p[i]].he=he[i];

      vhe[he[i]].opp=heo[i];
      vhe[heo[i]].opp=he[i];
    }
    for(int i=0;i<5;i++) {
      vhe[he[i]].next=he[(i+1)%5];
      vhe[heo[i]].point=p[(i+1)%5];
      vhe[heo[i]].next=heo[(i+4)%5];
    }
    if(th)
      for(int i=0;i<5;i++)
	th[i]=heo[i];
    makeprev();
  }

  static char nextangle(char x)
  {
    if(x>='a' && x<='z') return 'a'+(x-'a'+1)%5;
    if(x>='A' && x<='Z') return 'A'+(x-'A'+4)%5;
    assert(0);
  }

  static bool isspecial(char x) {
    if(x>='a' && x<='z') return 0;
    if(x>='A' && x<='Z') return 0;
    return 1;
  }

  static int aproj(char x)
  {
    if(x>='a' && x<='z') return x-'a';
    if(x>='A' && x<='Z') return x-'A';
    assert(0);
  }

  static char nextlen(char x)
  {
    if(x>='a' && x<='z') return x;
    if(x>='A' && x<='Z') return 'A'+(x-'A'+4)%5;
    assert(0);
  }

  char opplen(int i) const
  {
    int opp=vhe[i].opp;
    return nextlen(vhe[opp].angle);
  }

  int iopplen(int i) const {
    char c=opplen(i);
    return aproj(c);
  }

  void makeprev() {
    for(int i=0;i<vhe.size();i++)
      vhe[next(i)].prev=i;
  }

  int nexta(int he) const {
    return vhe[vhe[he].opp].next;
  }

  int next(int he) const {
    return vhe[he].next;
  }

  int prev(int he) const {
    int r=vhe[he].prev;
    assert(vhe[r].next==he);
    return r;
  }

  int opp(int he) const {
    return vhe[he].opp;
  }

  int getangle(int p,char a) const {
    int he=vpoint[p].he;
    int he2=he;
    int r=-1;
    do {
      if(vhe[he].angle==a) {
	assert(r<0);
	r=he;
      }
      he=nexta(he);
    } while(he2!=he);
    return r;
  }

  int getunk(int p) const {
    return getangle(p,ANG_UNK);
  }

  int get180(int p) const {
    return getangle(p,ANG_180);
  }

  /* "merge" */

  bool makenext(int u1,int u2,char a1=0,char a2=0) {   // fait que u2 soit le next de u1
    if(dbg_nt>1) printf("makenext %d %d %c %c\n",u1,u2,a1?a1:'/',a2?a2:'/');
    assert(u1!=u2);

    int v1=next(u1);
    int v2=prev(u2);
    int p1=vhe[v1].point;
    int p2=vhe[u2].point;

    assert(vhe[v1].angle==ANG_UNK && vhe[u2].angle==ANG_UNK);

    assert(p1!=p2);

    vhe[u1].next=u2;
    vhe[v2].next=v1;

    int he=u2;
    int he2=he;
    do {
      vhe[he].point=p2;
      he=nexta(he);
    } while(he2!=he);
    vpoint[p1].he=-1;
    freepoints.push_front(p1);
    if(a1) vhe[u2].angle=a1;
    if(a2) vhe[v1].angle=a2;
    makeprev();

    auto v=getpointv(p2);
    assert(v[5]<2 && v[6]==1);
    int r1=testsub5(v,vA1); //peut etre type1
    int r2=(v[5]==0)?testsub5(v,vA2):0; //peut etre type2

    if(r1==0 && r2==0) {
      if(dbg_nt) cout<<"makenext (end): vA1/2\n";
      return false;
    }
    if(r1==2)
      vhe[getunk(p2)].angle=(v[5]>0)?ANG_0:ANG_180;
    else if(r2==2)
      vhe[getunk(p2)].angle=ANG_0;

    return 1;
  }

  /* pos lines */

  struct pos_t {
    pos_t(int a=1,int b=1,int c=1) {
      eg=a;l=b;g=c;
      done=0;
    }
    bool isdead() const {
      if(eg||l||g) return 0;
      return 1;
    }
    bool iseg() const {
      if(l||g) return 0;
      return eg;
    }
    void imp(int a,int b,int c,int d=0) {
      if(a==0) eg=0;
      if(b==0) l=0;
      if(c==0) g=0;
      if(d) done=1;
    }
    char eg,l,g,done;
  };

  bool addl(const vi_t &v) {
    if(dbg_nt>2) ::aff(v,"\n","addl: ");
    solved=0;
    int d=cl3.dim();
    vector<rat_t> vv2(5);
    for(int i=0;i<5;i++)
      vv2[i]=v[i];
    cl3.cut(vv2,0);
    if(nocut && cl3.dim()!=d) return 0;
    if(cl3.empty()) return 0;
    if(!testsol()) return 0;
    return 1;
  }

  bool addeg(const vi_t &v) {
    if(dbg_nt>2) ::aff(v,"\n","addeg: ");
    solved=0;
    int d=cl3.dim();
    vector<rat_t> vv2(5);
    for(int i=0;i<5;i++)
      vv2[i]=v[i];
    cl3.proj(vv2,0);
    if(nocut && cl3.dim()!=d) return 0;
    if(cl3.empty()) return 0;
    if(!testsol()) return 0;
    return 1;
  }

  map<vi_t,pos_t> mpo;

  //-1 : dead / 0: nothing / 1: it's eg / 2: it's eg and new
  int impose(const vi_t &v,int eg,int l,int g) {
    if(dbg_nt>3) {printf("impose ");::aff(v,"");printf(" : %d %d %d \n",eg,l,g);}

    if(cl3.c.comp(v,0)) {
      l=g=0;
      pos_t &p(mpo[v]); //théoriquement, on pourrait ne pas mettre dans mpo;
      p.done=1;
      p.imp(eg,l,g);
      assert(eg);
      return 1;
    }

    pos_t &p(mpo[v]);
    p.imp(eg,l,g);
    if(p.isdead()) return -1;
    if(p.done==0 && p.eg==1 && p.l==0 && p.g==0) {
      //c'est egal, mais nouveau;
      int ra=addeg(v);
      if(ra==0) return -1;
      p.done=1;
      return 2;
    }
    return 0;
  }

  /* allposlines : returns 0 (ok), -1 (ko) or 1 (ok & redo)*/

  int allposlines_(int he,set<int> &_pl) {
    if(dbg_nt>3) printf("===\n");
    while(vhe[(he)].angle==ANG_0 || vhe[(he)].angle==ANG_180)
      he=next(he);

    struct line_t {
      int p1,p2,he1,he2,np;
      vi_t v[3];
      line_t(int p1_,int p2_,int np_,vi_t v_[3],int he1_,int he2_):p1(p1_),p2(p2_),he1(he1_),he2(he2_),np(np_) {
	memcpy(v,v_,sizeof(v));
      }
    };

    list<line_t> ps;
    {
      int p1=-1,p2=-1;
      string xx;
      int he2=he;
      int fp1=-1;
      int np=0;
      vi_t v[3];
      int sthe=-1,enhe=-1;
      while(1) {
	if(vhe[(he2)].angle!=ANG_0 && vhe[(he2)].angle!=ANG_180) {
	  assert(vhe[(he2)].angle==ANG_UNK);
	  p2=vhe[he2].point;
	  if(p1>=0) {
	    if(dbg_nt>3) printf("allposlines : %d(%d) %s %d(%d) \n",p1,sthe,xx.c_str(),p2,enhe);
	    if(np)
	      ps.push_back(line_t(p1,p2,np,v,sthe,enhe));
	  }
	  if(fp1>=0 && he2==he) break;
	  xx="";
	  np=0;
	  memset(v,0,sizeof(v));
	  p1=vhe[he2].point;
	  if(fp1<0) fp1=p1;
	  xx+=string(1,opplen(he2));
	  v[np][iopplen(he2)]++;
	  enhe=sthe=he2;
	} else {
	  if(vhe[(he2)].angle==ANG_0) {//?
	    xx+=string(1,'.');
	    np++;
	    if(np>2) {
	      //on pourrait essayer de faire un truc pour pas entrer là dedans...
	      if(dbg_nt) cout<<"allposline si.size() > 2 : impossible. Trouver contradiction avant ? si om branche sur un sommet de type2, il peut pas être au bout d'une chaine xx.xxx.xx\n";
	      return -1;
	    }
	  }
	  xx+=string(1,opplen(he2));
	  v[np][iopplen(he2)]++;
	  enhe=he2;
	}

	_pl.insert(he2);
	he2=next(he2);
      }
      assert(p2==fp1); //on a bien fait le tour
    }

    for(auto &jt:ps) {

      bool fin1=(testsub5(getpointv(jt.p1),vA1)==false);
      bool fin2=(testsub5(getpointv(jt.p2),vA1)==false);

      if(dbg_nt>3) printf("fins= %d %d\n",fin1,fin2);

      //tout ce qui importe, c'est de décider/savoir si p1 et p2 sont les mêmes points

      if(jt.np==1) {
	if(dbg_nt>3) printf("case1\n");
	int ri=impose(jt.v[0]-jt.v[1],1,!fin1,!fin2);
	if(dbg_nt>3) printf("ri=%d\n",ri);
	if(ri==-1)
	  return -1;
	if(ri) {
	  if(dbg_nt>1) printf("allposlines: fusion(1) %d %d\n",jt.p1,jt.p2);
	  if(!makenext(jt.he2,jt.he1,ANG_0)) return -1;
	  return 1;
	}
      } else {  //jt.np==2
	if(dbg_nt>3) printf("case2\n");
	if(fin1 || fin2) {
	  if(dbg_nt>1) printf("imp fins\n");
	  return -1;
	}
	int ri=impose(jt.v[0]+jt.v[2]-jt.v[1],1,1,0);
	if(dbg_nt>3) printf("ri=%d\n",ri);
	if(ri==-1)
	  return -1;
	if(ri) {
	  if(dbg_nt>1) printf("allposlines: fusion(2) %d %d\n",jt.p1,jt.p2);
	  if(!makenext(jt.he2,jt.he1,ANG_180)) return -1;
	  return 1;
	}

      }
    }
    return 0;
  }

  int allposlines() {
    if(dbg_nt) printf("allposlines\n");
    set<int> _pl;
    int re=0;
    for(int i=0;i<vhe.size();i++) {
      if(vhe[i].angle!=ANG_UNK) continue;
      if(_pl.find(i)!=_pl.end()) continue;
      if(dbg_nt) printf("allposlines(%d)\n",i);
      int r=allposlines_(i,_pl);
      if(dbg_nt) printf("allposlines(%d) ret: %d\n",i,r);
      if(r<0) {re=-1;break;}
      if(r) re=1;
    }
    return re;
  }

  /* les closes : fusionner ? */

  template<class v_t,class v2_t>
  static int testsub5(const v_t &v,const v2_t &vA) { //0 : mort// 1: pas mort sur subneq //2: exact
    for(auto &it:vA) {
      int r=2;
      for(int i=0;r && i<5;i++) {
	if(it[i]-v[i]!=0) r=1;
	if(it[i]-v[i]<0) r=0;
      }
      if(r) return r;
    }
    return 0;
  }

  vector<int> getpointv(int pt) const {
    int he=vpoint[pt].he;
    int h2=he;
    vector<int> a(7);
    do {
      char c=vhe[he].angle;
      if(c==ANG_0) {}
      else if(c==ANG_180) a[5]++;
      else if(c>='a' && c<='e') a[c-'a']++;
      else if(c>='A' && c<='E') a[c-'A']++;
      else a[6]++;
      he=nexta(he);
    } while(he!=h2);
    assert(a[6]<2 && a[5]<2);
    return a;
  }

  int close() {
    if(dbg_nt>1) printf("close\n");
    int r,re=0;
    while(1) {
      if(dbg_nt>1) printf("close...\n");
      r=allposlines();
      if(r<0) return r;
      if(r) {
	re=1;
	settaff(*this);
	pause("dans close, avant retry\n");
	continue;
      }
      return re;
    }
  }

  /* testsol (sortir de tile.hpp ??) */

  bool solved;   //si solved=1, c'est qu'on a trouvé des angles/longeurs qui satisfaissent les conditions
  vector<double> vs_,vs2_;

  int testsol(int recomp=0) {
    if(dbg_nt) { printf("=== TESTSOL ===\n"); mainca.aff(); cl3.aff(); }

    if(cl3.dim()<0) { //TODO added 3/4/17
      solved=0;
      return 0;
    }

    assert(cl3.dim()>=0);

    static map<cp_t,vector<double> > memo2; //memorisation

    auto it=memo2.find(cl3);
    if(recomp==0 && it!=memo2.end()) {
      if(dbg_nt) cout<<"sol in mem:\n";
      if(it->second.size()==0) {
	solved=0;
	if(dbg_nt) cout<<"no sol!\n";
      } else {
	solved=1;
	vs_.resize(5);
	vs2_.resize(5);
	for(int i=0;i<5;i++)
	  vs_[i]=it->second[i];
	for(int i=0;i<5;i++)
	  vs2_[i]=it->second[i+5];
	if(dbg_nt) { cout<<"sol:\n";::aff(vs_);::aff(vs2_);}
      }
      return solved;
    }
    vector<double> &mem(memo2[cl3]);
    assert(mem.size()==0);

    if(mainca.dim==0) {
      solved=cpan(mainca,cl3,vs_,vs2_);
      if(dbg_nt>1) printf("cpan returns %d\n",solved);
      if(solved) {
	mem.resize(10);
	for(int i=0;i<5;i++) {
	  mem[i]=vs_[i];
	  mem[i+5]=vs2_[i];
	}
      }
      return solved;
    }

    solved=hassol(mainca,cl3,vs_,vs2_);
    if(dbg_nt>1) printf("hassol returns %d\n",solved);
    if(solved) {
      mem.resize(10);
      for(int i=0;i<5;i++) {
	mem[i]=vs_[i];
	mem[i+5]=vs2_[i];
      }
    }
    return solved;
  }

  /* liste des branchements */

  struct branch2_t {
    char angle,a180;
    branch2_t(int a,int b):angle(a),a180(b) {}
    bool operator<(const branch2_t &b) const {if(a180!=b.a180) return a180>b.a180;return angle<b.angle;}
  };

  set<branch2_t> poscote2(const vector<int> &v) const {
    set<branch2_t> s;
    int n180=v[5];
    assert(n180<2);

    if(n180==0)
      for(auto &it:vA2) {
	vi_t b(v0);
	bool bad=0;
	for(int i=0;i<5;i++) {
	  b[i]=it[i]-v[i];
	  if(b[i]<0) bad=1;
	}
	if(bad) continue;
	bool ok=0;
	for(int i=0;i<5;i++)
	  if(b[i]) {
	    ok=1;
	    s.insert(branch2_t(i,0));
	  }
	assert(ok);
      }
    for(auto &it:vA1) {
      vi_t b(v0);
      bool bad=0;
      for(int i=0;i<5;i++) {
	b[i]=it[i]-v[i];
	if(b[i]<0) bad=1;
      }
      if(bad) continue;
      bool ok=0;
      if(n180==0) {
	ok=1;
	for(int i=0;i<5;i++)
	  if(b[i]) {
	    s.insert(branch2_t(i,0));
	    s.insert(branch2_t(i,1));
	    ok=1;
	  }
      } else {
	for(int i=0;i<5;i++)
	  if(b[i]) {
	    s.insert(branch2_t(i,0));
	    ok=1;
	  }
      }
      assert(ok);
    }
    return s;
  }

  /* ORACLE, pas important de vérifier */

  int nexttobranch() const {
    if(dbg_nt>1) printf("nexttobranch\n");
    uf_t uf;
    set<int> sf;
    set<int> _pl;
    _pl.clear();
    for(int i=0;i<vhe.size();i++) {
      if(vhe[i].angle!=ANG_UNK) continue;
      if(_pl.find(i)!=_pl.end()) continue;

      int he2=i,he3=i;
      int fp1=-1;
      int np=0;
      set<int> ss;
      while(1) {
	sf.insert(he2);
	if(vhe[(he2)].angle!=ANG_0 && vhe[(he2)].angle!=ANG_180) {
	  if(np>=algobnd) {
	    for(auto &it:ss) {
	      if(dbg_nt>2) printf("uf merge %d %d (np=%d)\n",he3,it,np);
	      uf.merge(he3,it);
	    }
	  }
	  if(fp1>=0 && he2==i) break;
	  if(fp1<0) fp1=0;
	  he3=he2;
	  np=0;
	  ss.clear();
	} else {
	  //printf("merge %d %d\n",he2,he3);
	  ss.insert(he2);
	  if(vhe[(he2)].angle==ANG_0) {np++;} //?
	}
	_pl.insert(he2);
	he2=next(he2);
      }
    }
    map<int,list<int> > rm;

    for(auto &it:sf) {
      rm[uf.get(it)].push_back(it);
    }

    map<int,int> sc;
    multimap<int,int> lsc;

    for(auto &it:rm) {
      int m=-1;
      for(auto &jt:it.second) {
	int f=vhe[opp(jt)].face;
	assert(f>=0);
	if(m<0 || m>f) m=f;
      }

      for(auto &jt:it.second) {
	int p=vhe[jt].point;
	int p2=vhe[opp(jt)].point;
	if(sc.find(p)==sc.end() || sc.at(p)>m) sc[p]=m;
	if(sc.find(p2)==sc.end() || sc.at(p2)>m) sc[p2]=m;
      }
    }

    for(auto &it:sc) {
      lsc.insert(make_pair(it.second,it.first));
      if(dbg_nt>2) printf("sc %d : %d\n",it.first,it.second);
    }

    for(auto &it:lsc)
      if(getunk(it.second)>=0)
	return it.second;

    assert(0);
    return -1;
  }

  /* COMP POSITIONS (in tile_disp.hpp) */

  void comppos(long double *a,long double *l,long double angle0=0);
  void comppos(double aff_angle0=0);
  void comppos_spr();
  void comppos_dospr(const set<int> &fix);

#ifdef CAIRO
  void aff(cairo_t *cr);
#endif

  /* FOR INFORMATION */

  void aff() const {
    for(int i=0;i<vhe.size();i++)
      printf("%d : p=%d f=%d opp=%d next=%d nexta=%d a=%c\n",i,vhe[i].point, vhe[i].face,opp(i),next(i),nexta(i),char(vhe[i].angle));
    for(int i=0;i<vpoint.size();i++)
      if(vpoint[i].he>=0)
	printf("P%d : he=%d info=%s\n",i,vpoint[i].he,pointinfo2(i).c_str());
  }

  string getpointstr(int pt) const { // seulement pour infos
    int he=vpoint[pt].he;
    int h2=he,h3=-1;
    do {
      if(vhe[he].angle==ANG_UNK) {
	h3=he;
	break;
      }
      he=nexta(he);
    } while(he!=h2);
    if(h3>=0)
      he=h3;
    h2=he;
    string r;
    do {
      r+=string(1,vhe[he].angle);
      he=nexta(he);
    } while(he!=h2);
    return r;
  }

  string pointinfo(int i) const {
    auto v=getpointv(i);
    assert(v[5]<2 && v[6]<2);
    int r1=testsub5(v,vA1); //peut etre type1
    int r2=(v[5]==0)?testsub5(v,vA2):0; //peut etre type2
    char ww[100];
    snprintf(ww,99,"r1=%d r2=%d",r1,r2);
    string r=ww;
    r+=string(" ")+pointinfo2(i);

    return r;
  }

  string pointinfo2(int i) const {
    string r;
    int he=vpoint[i].he;
    int he2=he;
    do {
      r+=string(1,vhe[he].angle);
      he=nexta(he);
    } while(he2!=he);
    string ss=getpointstr(i);
    return r+" "+ss+" ";
  }
};

