
//pour affichage/CAIRO

#ifdef CAIRO

#include <cairo.h>
#include <cairo-xlib.h>
#include <cairo-svg.h>
#include <cairo-ps.h>
#include <cairo-pdf.h>

#define W_X (600)
#define W_Y (600)

extern const char *display;

typedef struct win {
    Display *dpy;
    int scr;

    Window win;
    GC gc;

    int width, height;
    KeyCode quit_code;
} win_t;

extern win_t *win;

extern void win_handle_events(win_t *win);

extern void g_init();
extern void g_deinit();

extern void cairo_aff(const string &info="");
extern void cairo_aff_pdf(const string &info="");
double cpos[2]={0,0};
double cscale=800;

int wsizex=600;
int wsizey=600;

#define AFF_TEXT_POINT 3
#define AFF_DOT 8
#define AFF_ANGLE 16
#define AFF_TEXT_ANGLE 32
#define AFF_M_ANGLE (AFF_TEXT_ANGLE|AFF_ANGLE)
#define AFF_TEXT_HE 64
#define AFF_TEXT_INFO 128

int affinfo=AFF_M_ANGLE | AFF_TEXT_INFO;

double dista=1;

double aff_angle0=0;

void scale(double a,double b, double &c, double &d,int s,int ss)
{
  a-=cpos[0];
  b-=cpos[1];
  a*=cscale;
  b*=-cscale;
  a+=s/2;
  b+=ss/2;
  c=a;
  d=b;
}

void rscale(double a,double b, double &c, double &d,int s,int ss)
{
  a-=s/2;
  b-=ss/2;
  a/=cscale;
  b/=-cscale;
  a+=cpos[0];
  b+=cpos[1];
  c=a;
  d=b;
}

win_t winn;
win_t *win=&winn;

const char *display="";

static void
win_init(win_t *win)
{
    Window root;

    win->width = W_X;
    win->height = W_Y;

    root = DefaultRootWindow(win->dpy);
    win->scr = DefaultScreen(win->dpy);

    win->win = XCreateSimpleWindow(win->dpy, root, 0, 0,
				   win->width, win->height, 0,
				   BlackPixel(win->dpy, win->scr), BlackPixel(win->dpy, win->scr));

    win->quit_code = XKeysymToKeycode(win->dpy, XStringToKeysym("Q"));

    XSelectInput(win->dpy, win->win,
		 KeyPressMask
		 |StructureNotifyMask
		 |ExposureMask
		 |ButtonPressMask |ButtonReleaseMask | PointerMotionMask |Button1MotionMask|Button2MotionMask|Button3MotionMask
		 );

    XMapWindow(win->dpy, win->win);
}

static void
win_deinit(win_t *win)
{
    XDestroyWindow(win->dpy, win->win);
}

void g_init()
{
    winn.dpy = XOpenDisplay(display);

    if (winn.dpy == NULL) {
      fprintf(stderr, "Failed to open display %s\n",display);
	return;
    }

    win_init(win);
}

void g_deinit()
{
    win_deinit(win);

    XCloseDisplay(winn.dpy);
}

int cmx=0,cmy=0;
int modemove=0;
int modevmove=-1;

void t_t::comppos(long double *a,long double *l,long double angle0) {
  set<int> okp,okhe;
  {
    int p=0;
    for(;p<vpoint.size();p++)
      if(!isfreepoint(p)) break;
    if(p==vpoint.size()) return;
    okp.insert(p);
    vpoint[p].posx=0;
    vpoint[p].posy=0;
    int he=vpoint[p].he;
    vhe[he].posa=angle0;
    okhe.insert(he);
  }
  bool ok=1;
  do {
    ok=1;
    for(int he=0;he<vhe.size();he++) {
      if(okhe.find(he)!=okhe.end()) {
	if(okhe.find(opp(he))==okhe.end()) {
	  vhe[opp(he)].posa=vhe[he].posa+M_PI;
	  okhe.insert(opp(he));
	  ok=0;
	}
	int p=vhe[he].point;
	int p2=vhe[next(he)].point;
	int x=-1,y=-1;
	if(!isspecial(vhe[he].angle)) {
	  y=x=aproj(vhe[he].angle);
	  int s=vhe[he].fsens;
	  assert(s==1 || s==-1);
	  if(s==-1) x=(x+4)%5;
	  if(s==1) y=(y+1)%5;
	}
	if(x>=0) {
	  if(okp.find(p)!=okp.end() && okp.find(p2)==okp.end()) {
	    vpoint[p2].posx=vpoint[p].posx+cos(vhe[he].posa)*l[x];
	    vpoint[p2].posy=vpoint[p].posy+sin(vhe[he].posa)*l[x];
	    if(dbg_nt) printf("P %d\n",p2);
	    okp.insert(p2);
	    ok=0;
	  }
	  if(okp.find(p)==okp.end() && okp.find(p2)!=okp.end()) {
	    vpoint[p].posx=vpoint[p2].posx-cos(vhe[he].posa)*l[x];
	    vpoint[p].posy=vpoint[p2].posy-sin(vhe[he].posa)*l[x];
	    if(dbg_nt) printf("P %d\n",p);
	    okp.insert(p);
	    ok=0;
	  }
	}
      }
      int n=next(he);
      if(okhe.find(n)!=okhe.end() && okhe.find(he)==okhe.end()) {
	if(vhe[n].angle==ANG_0 ||vhe[n].angle==ANG_180 || !isspecial(vhe[n].angle)) {
	  double aa=0;
	  if(vhe[n].angle==ANG_180) aa=1;
	  if(!isspecial(vhe[n].angle)) aa=a[aproj(vhe[n].angle)];
	  vhe[he].posa=vhe[n].posa-M_PI+M_PI*aa;
	  if(dbg_nt) printf("HE %d %lf -> %d %lf \n",n,vhe[n].posa,he,vhe[he].posa);
	  okhe.insert(he);
	  ok=0;
	}
      }
      if(okhe.find(n)==okhe.end() && okhe.find(he)!=okhe.end()) {
	if(vhe[n].angle==ANG_0 || vhe[n].angle==ANG_180 ||!isspecial(vhe[n].angle)) {
	  double aa=0;
	  if(vhe[n].angle==ANG_180) aa=1;
	  if(!isspecial(vhe[n].angle)) aa=a[aproj(vhe[n].angle)];
	  vhe[n].posa=vhe[he].posa+M_PI-M_PI*aa;
	  if(dbg_nt) printf("HE %d %lf <- %d %lf \n",n,vhe[n].posa,he,vhe[he].posa);
	  okhe.insert(n);
	  ok=0;
	}
      }
    }
  } while(!ok);
}

void t_t::comppos(double aff_angle0) {
  int r1=testsol();
  if(r1==-1) {
    if(dbg_nt)printf("SKIP SOL\n");
    comppos_spr();
    return;
  }
  int r2=solved;
  assert(r1==r2);
  long double sol_a[5];
  long double sol_l[5];
  if(r1) {
    for(int i=0;i<5;i++) {
      sol_a[i]=vs_[i];
      sol_l[i]=vs2_[i];
    }
    if(dbg_nt) {printf("SOL:\n");::aff(vs_);::aff(vs2_);}
  } else {
    if(dbg_nt)printf("NO SOL\n");
    comppos_spr();
    return;
  }
  comppos(sol_a,sol_l,aff_angle0);
}

void t_t::comppos_dospr(const set<int> &fix) {
  for(int i=0;i<vpoint.size();i++) {
    if(isfreepoint(i)) continue;
    if(fix.find(i)!=fix.end()) continue;
    double px=0,py=0;
    int he=vpoint[i].he;
    int h2=he;
    int k=0;
    do {
      k++;
      int o=opp(he);
      int p2=vhe[o].point;
      px+=vpoint[p2].posx;
      py+=vpoint[p2].posy;
      he=nexta(he);
    } while(he!=h2);

    vpoint[i].posx=px/k;
    vpoint[i].posy=py/k;
  }
}

void t_t::comppos_spr() {
  for(int i=0;i<vpoint.size();i++) {
    if(isfreepoint(i)) continue;
    vpoint[i].posx=0;
    vpoint[i].posy=0;
  }

  vector<int> v;
  set<int> fix;
  for(int i=0;i<vhe.size();i++) {
    if(vhe[i].angle!=ANG_UNK) continue;

    int he2=i,he3=i;
    do {
      int p=vhe[he2].point;
      v.push_back(p);
      fix.insert(p);
      he2=next(he2);
    } while(he3!=he2);

    break;
  }
  assert(!v.empty());
  for(int i=0;i<v.size();i++) {
    int p=v[i];
    vpoint[p].posx=cos(2*M_PI/v.size()*i);
    vpoint[p].posy=-sin(2*M_PI/v.size()*i);
  }
  for(int i=0;i<100;i++)
    comppos_dospr(fix);
}

void t_t::aff(cairo_t *cr) {
  cairo_set_source_rgb(cr, 1,1,0);

  for(int i=0;i<this->vhe.size();i++) {
    double x,y,xx,yy;
    double x2,y2;
    double xd,yd;

    cairo_save(cr);

    int p=this->vhe[i].point;
    int p2=this->vhe[this->opp(i)].point;

    cairo_set_source_rgb(cr, 0,1,0);
    x=this->vpoint[this->vhe[i].point].posx;
    y=this->vpoint[this->vhe[i].point].posy;
    x2=5*x;y2=5*y;
    xd=x;yd=y;

    double xt=x*3;
    double yt=y*3;

    scale(x,y,xx,yy,wsizex,wsizey);
    cairo_move_to(cr, xx,yy);
    double xxp=xx,yyp=yy;

    x=this->vpoint[p2].posx;
    y=this->vpoint[p2].posy;
    xt+=x;
    yt+=y;
    xt/=4;
    yt/=4;

    x2+=x;y2+=y;
    x2/=6;y2/=6;
    xd-=x;yd-=y;
    double ra=atan2(-this->vpoint[p2].posy+this->vpoint[p].posy,this->vpoint[p2].posx-this->vpoint[p].posx);
    double dist=sqrt(xd*xd+yd*yd);

    double a=this->vhe[i].posa;
    scale(x,y,xx,yy,wsizex,wsizey);
    cairo_line_to(cr, xx,yy);

    cairo_stroke(cr);

    if(affinfo &AFF_TEXT_HE && !t_t::isspecial(this->vhe[i].angle)) {
      char w[100];
      snprintf(w,99,"%d %c %d",i,this->vhe[i].angle,this->vhe[i].face);
      scale(xt,yt,xx,yy,wsizex,wsizey);
      cairo_move_to(cr, xx+5,yy);
      cairo_set_font_size (cr, 16);
      cairo_show_text (cr,w);
    }

    cairo_restore(cr);

    scale(x2,y2,xx,yy,wsizex,wsizey);

    cairo_save(cr);
    cairo_set_source_rgb(cr, 1,1,1);
    cairo_set_font_size (cr, 16);
    cairo_move_to(cr,xx-40*dist*dista*sin(a)-8+rand()%16,yy-40*dist*dista*cos(a)+10+rand()%16);
    char w[100];
    cairo_stroke(cr);

    cairo_restore(cr);

    int ii=prev(i);
    int p3=this->vhe[ii].point;
    double ra2=atan2(-this->vpoint[p3].posy+this->vpoint[p].posy,this->vpoint[p3].posx-this->vpoint[p].posx);

    cairo_save(cr);
    if(fabs(ra2-ra)>0.01 && fabs(ra2-ra)<2*M_PI-0.01) {
      int ok=3;
      if(this->vhe[i].angle=='0')
	{cairo_set_source_rgb(cr, 0,1,0);ok=2;}
      else if(this->vhe[i].angle=='_')
	cairo_set_source_rgb(cr, 1,0,0);
      else if(this->vhe[i].angle=='?')
	{ok=1;cairo_set_source_rgb(cr, 0,0,1);}
      else cairo_set_source_rgb(cr, 1,1,1);
      if((affinfo& AFF_ANGLE) && (ok&2)) {
	cairo_set_line_width (cr, 1);
	if(this->vhe[i].angle=='0')
	  cairo_set_line_width (cr, 8);
	cairo_arc(cr,xxp,yyp, 30, ra2 , ra);
	cairo_stroke(cr);
      }
      if((affinfo&AFF_TEXT_ANGLE) && (ok&1)) {
	double ra3=(ra+ra2)/2;
	if(ra<ra2) ra+=2*M_PI;
	while(ra3>ra) ra3-=M_PI;
	while(ra3<ra2) ra3+=M_PI;
	cairo_set_font_size (cr, 16);
	cairo_move_to(cr,xxp+17*cos(ra3)-8+rand()%1,yyp+17*sin(ra3)+5+rand()%1);
	snprintf(w,100,"%c",this->vhe[i].angle);
	cairo_show_text (cr,w);
      }
    }
    cairo_restore(cr);

  }

  for(int i=0;i<this->vpoint.size();i++) {
    if(this->isfreepoint(i)) continue;
    double x,y,xx,yy;

    x=this->vpoint[i].posx;
    y=this->vpoint[i].posy;

    scale(x,y,xx,yy,wsizex,wsizey);

    if(affinfo&AFF_DOT) {
      cairo_set_line_width (cr, 1);
      cairo_set_source_rgb(cr, 1,1,1);
      cairo_save(cr);
      cairo_arc(cr,xx,yy, 4, 0, 2.0 * 3.1415);
      cairo_fill(cr);
      cairo_restore(cr);
    }

    cairo_save(cr);
    cairo_set_source_rgb(cr, 1,1,1);
    cairo_set_font_size (cr, 16);
    if((affinfo&AFF_TEXT_POINT)==1)
      cairo_move_to(cr,xx+10,yy+5);
    else
      cairo_move_to(cr,xx+40,yy+10);
    char w[100];
    snprintf(w,100,"%d %s",i,this->pointinfo(/*cond,*/i).c_str());
    char w2[100];
    snprintf(w2,100,"%d",i);
    if((affinfo&AFF_TEXT_POINT)==3)
      cairo_show_text (cr,w);
    if((affinfo&AFF_TEXT_POINT)==1 || (affinfo&AFF_TEXT_POINT)==2 )
      cairo_show_text (cr,w2);
    cairo_stroke(cr);

    cairo_restore(cr);

  }
}

bool cairo=1;

t_t taff;

int findv(double x,double y)
{
  double dd=0;
  int b=-1;
  for(int i=0;i<taff.vpoint.size();i++) {
    if(taff.isfreepoint(i)) continue;
    double dd2=pow(x-taff.vpoint[i].posx,2)+pow(y-taff.vpoint[i].posy,2);
    if(b<0 || dd2<dd) {
      b=i;
      dd=dd2;
    }
  }
  return b;
}

int handle_events(win_t *win, bool wait=1)
{
  int reaff=0;

  while (1) {
    if(!XPending(win->dpy) && wait==0)
      return 0;

    XEvent xev;
    XNextEvent(win->dpy, &xev);

    switch(xev.type) {

    case MotionNotify:
      {
	XMotionEvent *me=&xev.xmotion;

	if(modemove) {
	  double a,b,c,d;
	  rscale(cmx,cmy,a,b,wsizex,wsizey);
	  rscale(me->x,me->y,c,d,wsizex,wsizey);
	  cpos[0]+=a-c;
	  cpos[1]+=b-d;
	  cmx=me->x;cmy=me->y;
	  reaff=-1;
	}

	if(modevmove>=0) {
	  double a,b,c,d;
	  rscale(cmx,cmy,a,b,wsizex,wsizey);
	  rscale(me->x,me->y,c,d,wsizex,wsizey);
	  taff.vpoint[modevmove].posx-=a-c;
	  taff.vpoint[modevmove].posy-=b-d;
	  cmx=me->x;cmy=me->y;
	  reaff=-1;
	}

	break;
      }

    case ButtonPress:
      {
	XButtonEvent *be= &xev.xbutton;

	if(be->button==1) {
	  double xx,yy;
	  rscale(be->x,be->y,xx,yy,wsizex,wsizey);

	  cmx=be->x;cmy=be->y;
	  modevmove=findv(xx,yy);
	}

	if(be->button==3) {
	  cmx=be->x;cmy=be->y;
	  modemove=1;
	}

	if(be->button==4) {
	  cscale*=1.5;
	}
	if(be->button==5) {
	  cscale/=1.5;
	}

	reaff=-1;
	break;
      }

    case ButtonRelease:
      {
	XButtonEvent *be= &xev.xbutton;

	if(be->button==3)
	  modemove=0;

	if(be->button==1)
	  modevmove=-1;

	break;
      }

    case ConfigureNotify:
      {
	XConfigureEvent *cev = &xev.xconfigure;

	if(wsizex!=cev->width || wsizey!=cev->height) reaff=-1;
	wsizex=win->width = cev->width;
	wsizey=win->height = cev->height;
      }
      break;
    case Expose:
      {
	XExposeEvent *eev = &xev.xexpose;

	int k=0;
	XEvent xev2;
	while(XCheckTypedEvent(win->dpy, Expose,&xev2)) k++;

	if (eev->count == 0)
	  reaff=-1;
      }
      break;
    case KeyPress:
      {
	XKeyEvent *kev = &xev.xkey;

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("B")))
	  return 'B';

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("exclam"))) {
	  if(autonext==0) {
	    autonext=-1;
	    return 'B';
	  }
	  autonext=0;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("Q")))
	  exit(0);

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("R"))) {
	  int r=taff.testsol(1);
	  printf("testsol ret %d\n",r);
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("A"))) {
	  taff.aff();
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("T"))) {
	  int x=affinfo&AFF_TEXT_POINT;
	  x=(x+1)&AFF_TEXT_POINT;
	  affinfo=x|(affinfo&~AFF_TEXT_POINT);
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("H"))) {
	  affinfo=affinfo^AFF_TEXT_HE;
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("J"))) {
	  affinfo=affinfo^AFF_TEXT_INFO;
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("I"))) { //INFO
	  cout<<"A "<<mainca.lin()<<endl;
	  cout<<"B "<<taff.cl3.c.lin()<<endl;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("Y"))) {
	  int x=affinfo&AFF_M_ANGLE;
	  x=(x+AFF_ANGLE)&AFF_M_ANGLE;
	  affinfo=x|(affinfo&~AFF_M_ANGLE);
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("D"))) {
	  setalldbg(10);
	  mopt["inter"]=2;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("M"))) {
	  for(int i=0;i<taff.vpoint.size();i++) {
	    if(taff.isfreepoint(i)) continue;
	    taff.vpoint[i].posx+=double(rand()%100-50)/1000;
	    taff.vpoint[i].posy+=double(rand()%100-50)/1000;
	  }
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("S"))) {
	  taff.comppos_spr();
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("P"))) {
	  vector<double> vs,vs2;
	  int r=hassol(mainca,taff.cl3,vs,vs2);
	  printf("hassol ret %d\n",r);
	  if(r) {
	    ::aff(vs);
	    ::aff(vs2);
	    long double a[5],b[5];
	    for(int i=0;i<5;i++) {
	      a[i]=vs[i];
	      b[i]=vs2[i];
	    }
	    taff.comppos(a,b,aff_angle0);
	    reaff=1;
	  }
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("W"))) {
	  taff.comppos(aff_angle0);
	  reaff=1;
	}

	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("F2"))) {
	  skippause=100;
	  return 'B';
	}
	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("F3"))) {
	  skippause=1000;
	  return 'B';
	}
	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("F4"))) {
	  skippause=10000;
	  return 'B';
	}
	if(kev->keycode==XKeysymToKeycode(win->dpy, XStringToKeysym("F5"))) {
	  skippause=100000;
	  return 'B';
	}
      }
    default: {}
    }

    if(!XPending(win->dpy)) {
      if(reaff) cairo_aff("-");
      reaff=0;
    }
  }

  return 0;
}

void cairo_aff(cairo_t *cr,const string &info_="")
{
  taff.aff(cr);

  if(affinfo &AFF_TEXT_INFO) {
    cairo_set_source_rgb(cr, 1,1,1);
    cairo_set_font_size (cr, 20);
    cairo_move_to(cr,10,70);
    cairo_show_text (cr,cmd.c_str());
    cairo_stroke(cr);

    cairo_move_to(cr,10,100);
    char w[100];
    snprintf(w,100,"ca.d=%d cl.d=%d #rec=%Ld #rec2=%Ld",mainca.dim,taff.cl3.dim(),nbrec,nbrec2);
    cairo_show_text (cr,w);
    cairo_stroke(cr);

    cairo_move_to(cr,10,40);
    cairo_show_text (cr,info_.c_str());
    cairo_stroke(cr);

    auto ek=isknownB(mainca,taff.cl3.c);
    string type="type:";
    for(int i=0;i<32;i++)
      if(ek&(1LL<<i)){
	char bf[100];
	snprintf(bf,99," %d",i);
	type+=bf;
	if(ek&(1LL<<(i+32)))
	  type+="E";
      }

    cairo_move_to(cr,10,130);
    cairo_show_text (cr,type.c_str());

    ostringstream str;
    str<<"a:";
    mainca.aff(str);
    auto sa=mainca;
    sa.sum();
    str<<"s:";
    sa.aff(str);
    str<<"b:";
    taff.cl3.aff(str);
    string str2=str.str();
    char xx[str2.size()+1];
    strcpy(xx,str2.c_str());
    char *p=strtok(xx,"\n");
    int pos=160;
    while(p) {
      cairo_move_to(cr,10,pos);
      cairo_show_text (cr,p);
      pos+=30;
      p=strtok(NULL,"\n");
    }
    cairo_stroke(cr);
  }

}

void cairo_aff(const string &info)
{
  if(info!="-") {
    if(dbg_nt) cout<<"== AFF "<<info<<" =="<<endl;
  }
  static string info_;
  if(info!="-") info_=info;

  if(!cairo) return;

  Visual *visual = DefaultVisual(win->dpy, DefaultScreen (win->dpy));

  cairo_surface_t *surface = cairo_xlib_surface_create (win->dpy, win->win, visual,
				       win->width, win->height);

  cairo_t *cr = cairo_create(surface);

  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_rectangle (cr, 0, 0,  win->width, win->height);
  cairo_fill (cr);

  cairo_aff(cr,info_);

  cairo_destroy(cr);
  cairo_surface_flush (surface);
  cairo_surface_finish (surface);
  //cairo_surface_mark_dirty (surface);
  cairo_surface_destroy (surface);

  XSync(win->dpy,0);
}

void cairo_aff_pdf(const string &info)
{
  static int k=0;
  char fn[1000];
  snprintf(fn,99,"out_%s_%d_",info.c_str(),k++);
  strcat(fn,cmd.c_str());
  strcat(fn,".pdf");

  for(int i=0;fn[i];i++) {
    if(fn[i]==' ')
      fn[i]='_';
    if(fn[i]=='/')
      fn[i]='_';
  }

  int size=2048;

  int wsizex_=wsizex;
  int wsizey_=wsizey;
  wsizex=size;
  wsizey=size;

  printf("out pdf to %s\n",fn);

  cairo_surface_t *surface = cairo_pdf_surface_create (fn,size,size);

  cairo_t *cr = cairo_create(surface);

  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_rectangle (cr, 0, 0,  size,size);
  cairo_fill (cr);

  cairo_aff(cr,info);

  cairo_destroy(cr);

  wsizex=wsizex_;
  wsizey=wsizey_;

  cairo_surface_flush (surface);
  cairo_surface_finish (surface);
  //cairo_surface_mark_dirty (surface);
  cairo_surface_destroy (surface);
}

#endif
