#ifndef __U_HPP__
#define __U_HPP__

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <iostream>
#include <set>
#include <map>
#include <vector>
#include <array>
#include <algorithm>
#include <list>
#include <tuple>
#include <fstream>
#include <sstream>
#include <math.h>
#include <gmpxx.h>

typedef long long ll_t;
typedef mpq_class rat_t;

using namespace std;

string cmd;

double getd(const rat_t &a) {return a.get_d();}
bool isnull(const rat_t &a) {return a==0;}
rat_t abs(const rat_t &r) {return r<0 ? -r:r;}

mpz_class pgcd(const mpz_class &a,const mpz_class &b)
{
  if(a==0) return b;
  if(b==0) return a;
  if(a<0) return pgcd(-a,b);
  if(b<0) return pgcd(a,-b);
  if(a<b) return pgcd(b%a,a);
  return pgcd(a%b,b);
}

mpz_class ppcm(const mpz_class &a,const mpz_class &b)
{
  return a*b/pgcd(a,b);
}

template<class i_t>
void aff(const i_t &t,const char *e="\n",const char *b="",ostream &cout=std::cout)
{
  cout<<b;
  for(int i=0;i<t.size();i++)
    cout<<t[i]<<" ";
  cout<<e;
}

template<class i_t>
void affd(const i_t &t,const char *e="\n",const char *b="",ostream &cout=std::cout)
{
  cout<<b;
  for(int i=0;i<t.size();i++)
    cout<<getd(t[i])<<" ";
  cout<<e;
}

template<class i_t>
bool isnull(const vector<i_t> &v)
{
  for(int i=0;i<v.size();i++)
    if(v[i]!=0) return 0;
  return 1;
}

template<class i_t,long unsigned int n>
i_t sum(const array<i_t,n> &a)
{
  i_t r=0;
  for(int i=0;i<a.size();i++)
    r+=a[i];
  return r;
}

template<class x_t,class i_t,long unsigned int n>
i_t sumprod(const array<x_t,n> &a,const vector<i_t> &b,int m=-1)
{
  i_t r=0;
  if(m==-1) m=n;
  for(int i=0;i<m;i++)
    r+=a[i]*b[i];
  return r;
}

typedef array<int,5> vi_t;

vi_t operator-(const vi_t &a,const vi_t &b) {
  vi_t r;
  for(int i=0;i<5;i++)
    r[i]=a[i]-b[i];
  return r;
}

vi_t operator+(const vi_t &a,const vi_t &b) {
  vi_t r;
  for(int i=0;i<5;i++)
    r[i]=a[i]+b[i];
  return r;
}

vi_t operator-(const vi_t &a) {
  vi_t r;
  for(int i=0;i<5;i++)
    r[i]=-a[i];
  return r;
}

bool iseven(const vi_t &vi)
{
  for(int i=0;i<5;i++)
    if(vi[i]%2) return false;
  return true;
}

vi_t div2(const vi_t &vi)
{
  auto r=vi;
  for(int i=0;i<5;i++)
    r[i]/=2;
  return r;
}

typedef vector<array<int,5> > vvi_t;
vi_t v0={0,0,0,0,0};

int getint(const char *&p)
{
  int r=0;
  assert(p[0]>='0' && p[0]<='9');
  while(p[0]>='0' && p[0]<='9') {
    r*=10;
    r+=p[0]-'0';
    p++;
  }
  return r;
}

vvi_t readvvi(const char *bf)
{
  vvi_t vv;
  const char *p=bf;
  while(p[0]>='0' && p[0]<='9') {
    vi_t v;
    int k=0;
    for(int i=0;i<5;i++) {
      v[k++]=(getint(p));
      assert(p[0]==' ');
      p++;
    }
    p++;
    vv.push_back(v);
  }
  return vv;
}

int cperm[5]={1,2,3,4,0};
int mperm[5]={0,4,3,2,1};

vi_t perm(const vi_t &v,int *p)
{
  vi_t r(v0);
  for(int i=0;i<5;i++)
    r[i]=v[p[i]];
  return r;
}

vvi_t sort(const vvi_t &v)
{
  multiset<vi_t> s;
  for(auto &it:v)
    s.insert(it);
  vvi_t r;
  for(auto &it:s)
    r.push_back(it);
  return r;
}

vvi_t perm(const vvi_t &v,int *p)
{
  vvi_t v2;
  for(auto &it:v)
    v2.push_back(perm(it,p));

  return sort(v2);
}

vvi_t minrep(const vvi_t &v)
{
  vvi_t r=v,v2=v;
  for(int i=0;i<5;i++) {
    if(v2<r) r=v2;
    v2=perm(v2,cperm);
  }
  v2=perm(v2,mperm);
  for(int i=0;i<5;i++) {
    if(v2<r) r=v2;
    v2=perm(v2,cperm);
  }
  return r;
}

/* GESTION OF OPTIONS */

map<string,ll_t> mopt;
map<string,string> mopts;
set<string> usedopt;

void readopt(int &ac_, char **&av_)
{
  int ac=ac_;
  char **av=av_;
  for(int i=1;i<ac && av[i][0]=='-';i++) {
    char *x=strdup(av[i]+1);
    char *p=x;
    while(p[0]&& p[0]!='=') p++;
    if(p[0]=='=') {p[0]=0;p++;}
    if(p[0]) {
      mopt[x]=atoll(p);
      mopts[x]=(p);
    }
    else
      mopt[x]=1;
    printf("'%s' = '%s'\n",x,p);
    free(x);
    ac_--;av_++;
  }
}

template<class x_t>
void setopt(x_t &a,const char *st) {
  if(mopt.find(st)!=mopt.end()) {
    a=mopt.at(st);
    usedopt.insert(st);
  }
}

void warnopt() {
  for(auto &it:mopt)
    if(usedopt.find(it.first)==usedopt.end())
      printf("warn not used opt : '%s'\n",it.first.c_str());
}

struct uf_t { /* union find */
  map<int,int> m;

  void merge(int a,int b) {
    a=get(a);
    b=get(b);
    if(a!=b)
      m[a]=b;
  }

  int get(int a) const {
    while(m.find(a)!=m.end()) a=m.at(a);
    return a;
  }
};

/* rank computation */

template<class q_t>
struct prank_t { //compute the rank incrementally
  int a;
  vector<int> bb;
  vector<vector<q_t> > M;
  int rank() const {return M.size();}
  prank_t(int a_):a(a_) {
    M.resize(0);
  }
  void del() {
    assert(M.size()>0);
    M.resize(M.size()-1);
    bb.resize(M.size());
  }
  void aff() const {
    for(int i=0;i<M.size();i++){
      for(int j=0;j<a;j++)
	cout<<M[i][j]<<"  ";
      cout<<endl;
    }
    cout<<"===\n";
  }
  bool add_(int k) {
    for(int i=0;i<k;i++) {
      int l=bb[i];
      assert(M[i][l]==1);
      q_t mm=-M[k][l];
      for(int j=0;j<a;j++)
	M[k][j]+=mm*M[i][j];
    }
    int nn=-1;
    for(int i=a-1;i>=0;i--)
      if(M[k][i]!=q_t(0)) nn=i;
    //printf("k=%d nn=%d\n",k,nn);
    if(nn==-1) {
      M.resize(k);
      return 0;
    }
    bb.push_back(nn);
    q_t x=M[k][nn];
    for(int i=0;i<a;i++)
      M[k][i]/=x;
    return 1;
  }

  template<class v_t>
  bool add(const v_t &v_)
  {
    int k=M.size();
    M.resize(M.size()+1);
    M[k].resize(a);
    for(int i=0;i<a;i++)
      M[k][i]=v_[i];
    return add_(k);
  }
  template<class v_t>
  bool add(const v_t &v_,const q_t &b)
  {
    int k=M.size();
    M.resize(k+1);
    M[k].resize(a);
    for(int i=0;i<a-1;i++)
      M[k][i]=v_[i];
    M[k][a-1]=b;
    return add_(k);
  }

  template<class v_t>
  bool test(const v_t &v_,const q_t &b)
  {
    int k=M.size();
    M.resize(k+1);
    M[k].resize(a);
    for(int i=0;i<a-1;i++)
      M[k][i]=v_[i];
    M[k][a-1]=b;
    bool r=add_(k);
    if(r) del();
    return r;
  }
};

template<class q_t>
int mrank(const vector<vector<q_t> > &M,int a) // computes the rank of M. 'a' is the number of columns
{
  prank_t<q_t> pp(a);
  for(auto &it:M)
    pp.add(it);
  return pp.rank();
}

#endif
